/**************************************************
 * VRML 2.0 Parser
 * Copyright (C) 1996 Silicon Graphics, Inc.
 *
 * Author(s)    : Gavin Bell
 *                Daniel Woods (first port)
 **************************************************
 */

//
// The VrmlNodeType class is responsible for storing information about node
// or prototype types.
//


#include "vrmlnode.h"
#include <iostream.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>

#define NULL 0
//
// Static list of node types.
//
list<VrmlNodeType*> VrmlNodeType::typeList;

VrmlNodeType::VrmlNodeType(const char *nm)
{
	assert(nm != NULL);
	name = strdup(nm);
        isclass=0;
}

VrmlNodeType::VrmlNodeType(const char *nm, int ic)
{ assert(nm != NULL);
  name = strdup(nm);
  isclass=ic;
}



VrmlNodeType::~VrmlNodeType()
{
	free(name);

    // Free strings duplicated when fields/eventIns/eventOuts added:
    list<NameTypeRec*>::iterator i;

    for (i = eventIns.begin(); i != eventIns.end(); i++) {
    	NameTypeRec *r = *i;
    	free(r->name);
        delete r;
    }
    for (i = eventOuts.begin(); i != eventOuts.end(); i++) {
    	NameTypeRec *r = *i;
    	free(r->name);
        delete r;
    }
    for (i = fields.begin(); i != fields.end(); i++) {
    	NameTypeRec *r = *i;
    	free(r->name);
        delete r;
    }
}

void
VrmlNodeType::addToNameSpace(VrmlNodeType *_type)
{
	if (find(_type->getName()) !=  NULL) {
    	cerr << "PROTO " << _type->getName() << " already defined\n";
        return;
    }
    typeList.push_front(_type);
}


// One list is used to store all the node types.  Nested namespaces are
// separated by NULL elements.
// This isn't terribly efficient, but it is nice and simple.
//
void
VrmlNodeType::pushNameSpace()
{
#ifdef STED
	typeList.push_front( (VrmlNodeType*) NULL);
#else
	typeList.push_front(NULL);
#endif
}

void
VrmlNodeType::popNameSpace()
{
	// Remove everything up to and including the next NULL marker:
    list<VrmlNodeType*>::iterator i;
    for (i = typeList.begin(); i != typeList.end();) {
    	VrmlNodeType *nodeType = *i;
        ++i;
        typeList.pop_front();

        if (nodeType ==  NULL) {
            break;
        }
        else {
            // NOTE:  Instead of just deleting the VrmlNodeTypes, you will
            // probably want to reference count or garbage collect them, since
            // any nodes created as part of the PROTO implementation will
            // probably point back to their VrmlNodeType structure.
            delete nodeType;
        }
    }
}

VrmlNodeType *
VrmlNodeType::find(const char *_name)
{   
    // Look through the type stack:
    list<VrmlNodeType*>::iterator i;
    for (i = typeList.begin(); i != typeList.end(); i++) {
    	VrmlNodeType *nt = *i;
    	if (nt !=  NULL && strcmp(nt->getName(),_name) == 0) {
            return nt;
        }
    }
    return  NULL;
}

void
VrmlNodeType::addEventIn(const char *name, int type, const char* typestring)
{
	add(eventIns, name, type, typestring);
};
void
VrmlNodeType::addEventOut(const char *name, int type, const char* typestring)
{
	add(eventOuts, name, type, typestring);
};
void
VrmlNodeType::addField(const char *name, int type, const char* typestring,
                       ParseTreeNode* pt)
{
	add(fields, name, type, typestring, pt);
};
void
VrmlNodeType::addField(const char *name, int type, const char* typestring)
{
	add(fields, name, type, typestring);
};
void
VrmlNodeType::addExposedField(const char *name, int type, const char* typestring,
                              ParseTreeNode* pt)
{
    char tmp[1000];
    add(fields, name, type, typestring);
    sprintf(tmp, "set_%s", name);
    add(eventIns, tmp, type, typestring);
    sprintf(tmp, "%s_changed", name);
    add(eventOuts, tmp, type,typestring, pt);
};
void
VrmlNodeType::addExposedField(const char *name, int type, const char* typestring)
{
    char tmp[1000];
    add(fields, name, type, typestring);
    sprintf(tmp, "set_%s", name);
    add(eventIns, tmp, type, typestring );
    sprintf(tmp, "%s_changed", name);
    add(eventOuts, tmp, type, typestring );
};

void VrmlNodeType::addSuperclass(const char *name)
 { if (strcmp(name, "SFNode") == 0) return; 
   VrmlNodeType *t = find(name);
   if (t==NULL) 
     { cerr << "Superclass " << name << " not defined\n";
       return;
     }
   superClasses.push_front(t);
 }

void
VrmlNodeType::add(list<NameTypeRec*> &recs, const char *name, 
                  int type, const char* typestring)
{
    NameTypeRec *r = new NameTypeRec;
    r->name = strdup(name);
    r->type = type;
    r->typestring = strdup(typestring);
    r->pt=NULL;
    recs.push_front(r);
}

void
VrmlNodeType::add(list<NameTypeRec*> &recs, const char *name, int type,
                  const char* typestring, ParseTreeNode* pt)
{
    NameTypeRec *r = new NameTypeRec;
    r->name = strdup(name);
    r->type = type;
    r->typestring = strdup(typestring);
    r->pt=pt;
    recs.push_front(r);
}

int
VrmlNodeType::hasEventIn(const char *name) const
{ int hf=has(eventIns, name);
  if ((!hf) && (isClass()))
    { list<VrmlNodeType*>::const_iterator j;
      for (j = superClasses.begin(); j != superClasses.end(); j++) 
      { hf=(*j)->hasEventIn(name);
        if (hf) return(hf);
      }
    }
  return hf;
}
int
VrmlNodeType::hasEventOut(const char *name) const
{ int hf=has(eventOuts, name);
  if ((!hf) && (isClass()))
    { list<VrmlNodeType*>::const_iterator j;
      for (j = superClasses.begin(); j != superClasses.end(); j++) 
      { hf=(*j)->hasEventOut(name);
        if (hf) return(hf);
      }
    }
  return hf;
}
int
VrmlNodeType::hasField(const char *name) const
{ int hf=has(fields, name);
  if ((!hf) && (isClass()))
    { list<VrmlNodeType*>::const_iterator j;
      for (j = superClasses.begin(); j != superClasses.end(); j++) 
      { hf=(*j)->hasField(name);
        if (hf) return(hf);
      }
    }
  return hf;
}
int
VrmlNodeType::hasExposedField(const char *name) const
{
    // Must have field "name", eventIn "set_name", and eventOut
    // "name_changed", all with same type:
	char tmp[1000];
	int type;
	if ( (type = hasField(name)) == 0) return 0;
  
    sprintf(tmp, "set_%s\n", name);
    if (type != hasEventIn(name)) return 0;

    sprintf(tmp, "%s_changed", name);
    if (type != hasEventOut(name)) return 0;

    return type;
}

int
VrmlNodeType::has(const list<NameTypeRec*> &recs, const char *name) const
{
   list<NameTypeRec*>::const_iterator i;
   for (i = recs.begin(); i != recs.end(); i++) {
   if (strcmp((*i)->name, name) == 0)
       	return (*i)->type;
    }
    
   return 0;
}

VrmlNodeType::NameTypeRec*
VrmlNodeType::get(const list<NameTypeRec*> &recs, const char *name) const
{
   list<NameTypeRec*>::const_iterator i;
   for (i = recs.begin(); i != recs.end(); i++) {
   if (strcmp((*i)->name, name) == 0)
       	return (*i);
    }
    
   return NULL;
}



void 
VrmlNodeType::print_types()
{ 
  list<VrmlNodeType*>::iterator i;
  cerr << "\nPrototypes:\n";
  for (i = typeList.begin(); 
       i != typeList.end(); i++) {
       VrmlNodeType* nt=*i;
       if ((nt!=NULL) && (! nt->isClass()))
         cerr << nt->getName() << "\n";
    }
  cerr << "\nClasses:\n";
  for (i = typeList.begin(); 
       i != typeList.end(); i++) {
       VrmlNodeType* nt=*i;
       if ((nt!=NULL) && (nt->isClass()))
         cerr << nt->getName() << "\n";
    }
}