/**************************************************
 * VRML 2.0 Parser
 * Copyright (C) 1996 Silicon Graphics, Inc.
 *
 * Author(s)    : Gavin Bell
 *                Daniel Woods (first port)
 **************************************************
 */

//
// The VrmlNodeType class is responsible for storing information about node
// or prototype types.
//

#include "parstree.h"
#include <list.h>

#ifdef USE_STD_NAMESPACE
using namespace std;
#endif


class VrmlNodeType {
  public:
  	// Constructor.  Takes name of new type (e.g. "Transform" or "Box")
    // Copies the string given as name.
    VrmlNodeType(const char *nm);
    VrmlNodeType(const char *nm, int);

    // Destructor exists mainly to deallocate storage for name
    ~VrmlNodeType();

    // Namespace management functions.  PROTO definitions add node types
    // to the namespace.  PROTO implementations are a separate node
    // namespace, and require that any nested PROTOs NOT be available
    // outside the PROTO implementation.
    // addToNameSpace will print an error to stderr if the given type
    // is already defined.
    static void addToNameSpace(VrmlNodeType *);
    static void pushNameSpace();
    static void popNameSpace();

    // Find a node type, given its name.  
    // Returns NULL if type is not defined.
    static VrmlNodeType *find(const char *nm);

    typedef struct {
 	char *name;
        int type;
        char *typestring;
        ParseTreeNode* pt;
    } NameTypeRec;
        

    

    // Routines for adding/getting eventIns/Outs/fields 
    void addEventIn(const char *name, int type, const char* typestr);
    void addEventOut(const char *name, int type, const char* typestr);
    void addField(const char *name, int type, const char* typestr);
    void addExposedField(const char *name, int type, const char* typestr);
    void addField(const char *name, int type, const char* typestr, 
                  ParseTreeNode* pt);
    void addExposedField(const char *name, int type, const char* typestr,
                         ParseTreeNode* pt);
    void addSuperclass(const char *name);

    int hasEventIn(const char *name) const;
    int hasEventOut(const char *name) const;
    int hasField(const char *name) const;
    int hasExposedField(const char *name) const;
    
    int isClass() const { return isclass; };
    list<VrmlNodeType*> getSuperClasses() { return superClasses; }
    list<NameTypeRec*> getEventIns() { return eventIns; }
    list<NameTypeRec*> getEventOuts() { return eventOuts; }
    list<NameTypeRec*> getFields() { return fields; }

    const char *getName() const { return name; }

    static void print_types();

    NameTypeRec* VrmlNodeType::get(const list<NameTypeRec*> &,
                                   const char *name) const;
  private:
    void add(list<NameTypeRec*> &,const char *,int, const char*);
    void add(list<NameTypeRec*> &,const char *,int, const char*, ParseTreeNode*);
    int has(const list<NameTypeRec*> &,const char *) const;

    char *name;
    int isclass;

    // Node types are stored in this data structure:
    static list<VrmlNodeType*> typeList;

    list<VrmlNodeType*> superClasses;
    list<NameTypeRec*> eventIns;
    list<NameTypeRec*> eventOuts;
    list<NameTypeRec*> fields;
};

extern char* selfname;
